// Main.cpp

#include <iostream>

#include <Kinect.h>
#include <opencv2/opencv.hpp>
#include "cv_lib_macro.h"
#pragma CV_LIBRARY(world)

IKinectSensor *kinect_sensor = nullptr;

IColorFrameSource *color_frame_source = nullptr;
IColorFrameReader *color_frame_reader = nullptr;
IFrameDescription *color_frame_description = nullptr;

int color_width;
int color_height;

void ShowCapabilities(const DWORD cap)
{
  std::cout << "*** Kinect Capabilities: " << std::endl;
  std::cout << "Vision: "
            << ((cap & KinectCapabilities_Vision) ? "yes" : "no")
            << std::endl;
  std::cout << "Audio: "
            << ((cap & KinectCapabilities_Audio) ? "yes" : "no")
            << std::endl;
  std::cout << "Face: "
            << ((cap & KinectCapabilities_Face) ? "yes" : "no")
            << std::endl;
  std::cout << "Expressions: "
            << ((cap & KinectCapabilities_Expressions) ? "yes" : "no")
            << std::endl;
  std::cout << "Gamechat: "
            << ((cap & KinectCapabilities_Gamechat) ? "yes" : "no")
            << std::endl;
}

void ShowFrameDescription(IFrameDescription *fd)
{
  int width, height;
  unsigned int lip, bpp;
  float hfov, vfov, dfov;

  fd->get_Width(&width);
  fd->get_Height(&height);
  fd->get_LengthInPixels(&lip);
  
  fd->get_BytesPerPixel(&bpp);
  
  fd->get_HorizontalFieldOfView(&hfov);
  fd->get_VerticalFieldOfView(&vfov);
  fd->get_DiagonalFieldOfView(&dfov);

  std::cout << "size: width = " << width
            << ", height = " << height
            << ", total " << lip << " pixels" << std::endl;
  std::cout << "bytes per pixel = " << bpp << std::endl;
  std::cout << "field of view: horizontal = " << hfov
            << ", vertial = " << vfov
            << ", diagonal = " << dfov << std::endl;
}

HRESULT InitializeDefaultSensor(void)
{
    HRESULT hr;

    if (kinect_sensor) {
      std::cerr << "InitializeDefaultSensor(): already initialized"
                << std::endl;
    }

    hr = GetDefaultKinectSensor(&kinect_sensor);
    if (FAILED(hr)) {
      std::cerr << "GetDefaultKinectSensor() failed" << std::endl;
      return hr;
    }

    hr = kinect_sensor->Open();
    if (FAILED(hr)) {
      std::cerr << "IKinectSensor::Open() failed" << std::endl;
      return hr;
    }

    DWORD capabilities;
    hr = kinect_sensor->get_KinectCapabilities(&capabilities);
    if (FAILED(hr)) {
      std::cerr << "IKinectSensor::get_KinectCapabilities() failed"
                << std::endl;
      return hr;
    }
    ShowCapabilities(capabilities);

    // color

    hr = kinect_sensor->get_ColorFrameSource(&color_frame_source);
    if (FAILED(hr)) {
      std::cerr << "IKinectSensor::get_ColorFrameSource() failed" << std::endl;
      return hr;
    }

    hr = color_frame_source->OpenReader(&color_frame_reader);
    if (FAILED(hr)) {
      std::cerr << "IColorFrameSource::OpenReader() failed" << std::endl;
      return hr;
    }

    hr = color_frame_source->get_FrameDescription(&color_frame_description);
    if (FAILED(hr)) {
      std::cerr << "IColorFrameSource::get_FrameDescription() failed"
                << std::endl;
      return hr;
    }
    color_frame_description->get_Width(&color_width);
    color_frame_description->get_Height(&color_height);

    std::cout << "*** Color Frame:" << std::endl;
    ShowFrameDescription(color_frame_description);

    return hr;
}

void CloseSensor(void)
{
  if (color_frame_description) {
    color_frame_description->Release();
  }
  color_frame_description = nullptr;

  if (color_frame_reader) {
    color_frame_reader->Release();
  }
  color_frame_reader = nullptr;

  if (color_frame_source) {
    color_frame_source->Release();
  }
  color_frame_source = nullptr;
  
  if (kinect_sensor) {
    kinect_sensor->Close();
  }
  kinect_sensor = nullptr;
}

int main(int argc, char **argv)
{
  // initialiize kinect
  
  InitializeDefaultSensor();

  // initialize OpenCV

  char *window_color = "color";
  cv::namedWindow(window_color, CV_WINDOW_AUTOSIZE);

  cv::Size color_size(color_width, color_height);
  cv::Mat latest_color_image(color_size, CV_8UC4);

  // main loop

  while (1) {
    // color

    IColorFrame *color_frame = nullptr;
    HRESULT hr = color_frame_reader->AcquireLatestFrame(&color_frame);
    if (SUCCEEDED(hr)) {
      const int buf_size = sizeof(unsigned char)
          * color_width * color_height * 4;
      if (latest_color_image.isContinuous()) {
        hr = color_frame->CopyConvertedFrameDataToArray(buf_size,
            reinterpret_cast<BYTE *>(latest_color_image.data),
            ColorImageFormat_Bgra);
      } else {
        unsigned char *tmp_buf = new unsigned char[buf_size];
        hr = color_frame->CopyConvertedFrameDataToArray(buf_size, tmp_buf,
                                                        ColorImageFormat_Bgra);
        for (int y = 0; y < color_height; y++) {
          int line_size = sizeof(unsigned char) * color_width * 4;
          memcpy(latest_color_image.ptr<uchar>(y),
                 tmp_buf + line_size * y, line_size);
        }
        delete [] tmp_buf;
      }
      color_frame->Release();
    }

    cv::imshow(window_color, latest_color_image);

    int key = cv::waitKey(10);
    if (key == 3 || key == 27 || key == 'q') {
      break;
    }
  }

  CloseSensor();
 
  return 0;
}
