// Main.cpp

#include <iostream>

#include <Kinect.h>
#include <opencv2/opencv.hpp>
#include "cv_lib_macro.h"
#pragma CV_LIBRARY(world)

IKinectSensor *kinect_sensor = nullptr;

IColorFrameSource *color_frame_source = nullptr;
IColorFrameReader *color_frame_reader = nullptr;
IFrameDescription *color_frame_description = nullptr;

int color_width;
int color_height;

IDepthFrameSource *depth_frame_source = nullptr;
IDepthFrameReader *depth_frame_reader = nullptr;
IFrameDescription *depth_frame_description = nullptr;

int depth_width;
int depth_height;

void ShowCapabilities(const DWORD cap)
{
  std::cout << "*** Kinect Capabilities: " << std::endl;
  std::cout << "Vision: "
            << ((cap & KinectCapabilities_Vision) ? "yes" : "no")
            << std::endl;
  std::cout << "Audio: "
            << ((cap & KinectCapabilities_Audio) ? "yes" : "no")
            << std::endl;
  std::cout << "Face: "
            << ((cap & KinectCapabilities_Face) ? "yes" : "no")
            << std::endl;
  std::cout << "Expressions: "
            << ((cap & KinectCapabilities_Expressions) ? "yes" : "no")
            << std::endl;
  std::cout << "Gamechat: "
            << ((cap & KinectCapabilities_Gamechat) ? "yes" : "no")
            << std::endl;
}

void ShowFrameDescription(IFrameDescription *fd)
{
  int width, height;
  unsigned int lip, bpp;
  float hfov, vfov, dfov;

  fd->get_Width(&width);
  fd->get_Height(&height);
  fd->get_LengthInPixels(&lip);
  
  fd->get_BytesPerPixel(&bpp);
  
  fd->get_HorizontalFieldOfView(&hfov);
  fd->get_VerticalFieldOfView(&vfov);
  fd->get_DiagonalFieldOfView(&dfov);

  std::cout << "size: width = " << width
            << ", height = " << height
            << ", total " << lip << " pixels" << std::endl;
  std::cout << "bytes per pixel = " << bpp << std::endl;
  std::cout << "field of view: horizontal = " << hfov
            << ", vertial = " << vfov
            << ", diagonal = " << dfov << std::endl;
}

HRESULT InitializeDefaultSensor(void)
{
    HRESULT hr;

    if (kinect_sensor) {
      std::cerr << "InitializeDefaultSensor(): already initialized"
                << std::endl;
    }

    hr = GetDefaultKinectSensor(&kinect_sensor);
    if (FAILED(hr)) {
      std::cerr << "GetDefaultKinectSensor() failed" << std::endl;
      return hr;
    }

    hr = kinect_sensor->Open();
    if (FAILED(hr)) {
      std::cerr << "IKinectSensor::Open() failed" << std::endl;
      return hr;
    }

    DWORD capabilities;
    hr = kinect_sensor->get_KinectCapabilities(&capabilities);
    if (FAILED(hr)) {
      std::cerr << "IKinectSensor::get_KinectCapabilities() failed"
                << std::endl;
      return hr;
    }
    ShowCapabilities(capabilities);

    // color

    hr = kinect_sensor->get_ColorFrameSource(&color_frame_source);
    if (FAILED(hr)) {
      std::cerr << "IKinectSensor::get_ColorFrameSource() failed" << std::endl;
      return hr;
    }

    hr = color_frame_source->OpenReader(&color_frame_reader);
    if (FAILED(hr)) {
      std::cerr << "IColorFrameSource::OpenReader() failed" << std::endl;
      return hr;
    }

    hr = color_frame_source->get_FrameDescription(&color_frame_description);
    if (FAILED(hr)) {
      std::cerr << "IColorFrameSource::get_FrameDescription() failed"
                << std::endl;
      return hr;
    }
    color_frame_description->get_Width(&color_width);
    color_frame_description->get_Height(&color_height);

    std::cout << "*** Color Frame:" << std::endl;
    ShowFrameDescription(color_frame_description);

    // depth

    hr = kinect_sensor->get_DepthFrameSource(&depth_frame_source);
    if (FAILED(hr)) {
      std::cerr << "IKinectSensor::get_DepthFrameSource() failed" << std::endl;
      return hr;
    }

    hr = depth_frame_source->OpenReader(&depth_frame_reader);
    if (FAILED(hr)) {
      std::cerr << "IDepthFrameSource::OpenReader() failed" << std::endl;
      return hr;
    }

    hr = depth_frame_source->get_FrameDescription(&depth_frame_description);
    if (FAILED(hr)) {
      std::cerr << "IDepthFrameSource::get_FrameDescription() failed"
                << std::endl;
      return hr;
    }
    depth_frame_description->get_Width(&depth_width);
    depth_frame_description->get_Height(&depth_height);

    std::cout << "*** Depth Frame:" << std::endl;
    ShowFrameDescription(depth_frame_description);

    return hr;
}

void CloseSensor(void)
{
  if (depth_frame_description) {
    depth_frame_description->Release();
  }
  depth_frame_description = nullptr;

  if (depth_frame_reader) {
    depth_frame_reader->Release();
  }
  depth_frame_reader = nullptr;

  if (depth_frame_source) {
    depth_frame_source->Release();
  }
  depth_frame_source = nullptr;
  
  if (color_frame_description) {
    color_frame_description->Release();
  }
  color_frame_description = nullptr;

  if (color_frame_reader) {
    color_frame_reader->Release();
  }
  color_frame_reader = nullptr;

  if (color_frame_source) {
    color_frame_source->Release();
  }
  color_frame_source = nullptr;
  
  if (kinect_sensor) {
    kinect_sensor->Close();
  }
  kinect_sensor = nullptr;
}

int main(int argc, char **argv)
{
  // initialiize kinect
  
  InitializeDefaultSensor();

  // initialize OpenCV

  char *window_color = "color";
  cv::namedWindow(window_color, CV_WINDOW_AUTOSIZE);

  char *window_depth = "depth";
  cv::namedWindow(window_depth, CV_WINDOW_AUTOSIZE);

  cv::Size color_size(color_width, color_height);
  cv::Mat latest_color_image(color_size, CV_8UC4);

  cv::Size depth_size(depth_width, depth_height);
  cv::Mat latest_depth_image(depth_size, CV_16UC1);
  cv::Mat depth_result_channel[3];
  for (int i = 0; i < 3; i++) {
    depth_result_channel[i].create(depth_size, CV_8UC1);
  }
  cv::Mat depth_result_hsv_image(depth_size, CV_8UC3);
  cv::Mat depth_result_image(depth_size, CV_8UC3);

  // main loop

  while (1) {
    // color

    IColorFrame *color_frame = nullptr;
    HRESULT hr = color_frame_reader->AcquireLatestFrame(&color_frame);
    if (SUCCEEDED(hr)) {
      const int buf_size = sizeof(unsigned char)
          * color_width * color_height * 4;
      if (latest_color_image.isContinuous()) {
        hr = color_frame->CopyConvertedFrameDataToArray(buf_size,
            reinterpret_cast<BYTE *>(latest_color_image.data),
            ColorImageFormat_Bgra);
      } else {
        unsigned char *tmp_buf = new unsigned char[buf_size];
        hr = color_frame->CopyConvertedFrameDataToArray(buf_size, tmp_buf,
                                                        ColorImageFormat_Bgra);
        for (int y = 0; y < color_height; y++) {
          int line_size = sizeof(unsigned char) * color_width * 4;
          memcpy(latest_color_image.ptr<uchar>(y),
                 tmp_buf + line_size * y, line_size);
        }
        delete [] tmp_buf;
      }
      color_frame->Release();
    }

    // depth

    IDepthFrame *depth_frame = nullptr;
    hr = depth_frame_reader->AcquireLatestFrame(&depth_frame);
    if (SUCCEEDED(hr)) {
      unsigned short depth_min_reliable, depth_max_reliable;
      depth_frame->get_DepthMinReliableDistance(&depth_min_reliable);
      depth_frame->get_DepthMaxReliableDistance(&depth_max_reliable);

      const unsigned short depth_min = 0;
      const unsigned short depth_max = 8000;
      const unsigned short depth_range = depth_max - depth_min;

      unsigned int buffer_size;
      UINT16 *buffer;
      depth_frame->AccessUnderlyingBuffer(&buffer_size, &buffer);
      buffer_size *= sizeof(UINT16); // the manual says it's in bytes, but not.
      if (latest_depth_image.isContinuous()) {
        memcpy(latest_depth_image.data, buffer, buffer_size);
      } else {
        for (int y = 0; y < color_height; y++) {
          int line_size = sizeof(UINT16) * depth_width;
          memcpy(latest_depth_image.ptr<ushort>(y),
                 buffer + line_size * y, line_size);
        }
      }
      depth_frame->Release();

      // check min/max distance
#if 0      
      double min_value, max_value;
      cv::minMaxLoc(latest_depth_image, &min_value, &max_value);
      std::cout << "depth min = " << min_value << ", max = " << max_value
                << std::endl;
#endif
      // convert depth to image
#if 0
      const double a = -256.0 / depth_range;
      const double b = 256.0 * depth_max / depth_range;
      latest_depth_image.convertTo(depth_result_image, CV_8UC3, a, b);
#endif
#if 1
      const double a = 120.0 / depth_range;
      const double b = -120.0 * depth_min / depth_range;
      latest_depth_image.convertTo(depth_result_channel[0], CV_8UC1, a, b);
      depth_result_channel[1].setTo(192);
      latest_depth_image.convertTo(depth_result_channel[2], CV_8UC1);
      cv::threshold(depth_result_channel[2], depth_result_channel[2], 0, 255,
                    CV_THRESH_BINARY);
      cv::merge(depth_result_channel, 3, depth_result_hsv_image);
      cv::cvtColor(depth_result_hsv_image, depth_result_image, CV_HSV2BGR);
#endif      
    }

    cv::imshow(window_color, latest_color_image);
    cv::imshow(window_depth, depth_result_image);

    int key = cv::waitKey(10);
    if (key == 3 || key == 27 || key == 'q') {
      break;
    }
  }

  CloseSensor();
 
  return 0;
}
